@@ -243,7 +243,7 @@ End
       row[0] if row
     end
 
-    def thread(threads, depth, current, siblings, parents, msg_id_hash)
+    def thread(forest, depth, current, siblings, parents, msg_id_hash)
       result = Array.new
       return result if msg_id_hash.key?(current)
       msg_id_hash[current] = false
@@ -258,13 +258,13 @@ End
                                  :xref => xref,
                                  :nov => nov,
                                  :parents => parents | [numbers[0]]})
-      children = threads.select {|t| t[0] == current}.collect {|r| r[1]}
-      unless children.empty?
-        result.concat(thread(threads,depth.succ, children.shift, children,
+      children = forest[current]
+      if children
+        result.concat(thread(forest, depth.succ, children.shift, children,
                              parents | [numbers[0]], msg_id_hash))
       end
       siblings.each_with_index do |sibling, index|
-        result.concat(thread(threads,
+        result.concat(thread(forest,
                              depth, sibling, [], parents, msg_id_hash))
       end
       result
@@ -278,21 +278,19 @@ End
 SELECT number FROM disjoint WHERE parent = ?
 End
       result = Array.new
-      number_hash = Hash.new
-      numbers.each do |number|
-        next if number_hash.key?(number)
+      numbers = numbers.dup
+      until numbers.empty?
+        number = numbers.shift
         parent_stmt.execute(number)
         parent = parent_stmt.fetch[0].to_i
-        next if number_hash.key?(parent)
         number_stmt.execute(parent)
-        disjoint = number_stmt.fetch_all.collect {|row|
-          r = row[0].to_i
-          number_hash[r] = false
-          r
-        }
-        disjoint.delete_at(disjoint.index(parent))
-        disjoint.unshift(parent)
-        result << disjoint
+        set = number_stmt.fetch_all.collect {|row| row[0].to_i}
+        if set.length > 2
+          set[set.index(parent)] = set[0]
+          set[0] = parent
+        end
+        result << set
+        numbers -= set
       end
       result
     end
@@ -328,22 +326,31 @@ End
       result
     end
 
+    def thread_forest(sets)
+      forest = Hash.new
+      sets.each do |set|
+        set.each do |number|
+          row = @adapter.select_one(<<'End', number)
+SELECT msg_id FROM message WHERE number = ?
+End
+          parent = parent_msg_id(row[0])
+          forest[parent] ||= Array.new
+          forest[parent] << row[0]
+        end
+      end
+      forest
+    end
+
     def recent_threads(folder, length)
       rows = @adapter.select_all(<<'End', folder, length)
 SELECT number FROM filter WHERE folder = ? ORDER BY folder_number DESC LIMIT ?
 End
+      sets = disjoint_sets(rows.collect {|row| row[0].to_i})
+      forest = thread_forest(sets)
       result = Array.new
       msg_id_hash = Hash.new
-      threads = disjoint_sets(rows.collect {|row| row[0].to_i}).flatten.collect do |number|
-        row = @adapter.select_one(<<'End', number)
-SELECT msg_id FROM message WHERE number = ?
-End
-        [parent_msg_id(row[0]), row[0]]
-      end
-      threads.each do |t|
-        if t[0] == nil
-          result.concat(thread(threads, 0, t[1], [], [], msg_id_hash))
-        end
+      forest[nil].each do |root|
+        result.concat(thread(forest, 0, root, [], [], msg_id_hash))
       end
       result
     end
@@ -355,39 +362,28 @@ SELECT message.msg_id
   WHERE filter.folder = ? AND filter.folder_number >= ? AND filter.folder_number <= ?
   ORDER BY filter.folder_number ASC
 End
+      forest = Hash.new
+      rows.each do |row|
+        parent = parent_msg_id(row[0])
+        forest[parent] ||= Array.new
+        forest[parent] << row[0]
+      end
       result = Array.new
       msg_id_hash = Hash.new
-      threads = rows.collect do |row|
-        [parent_msg_id(row[0]), row[0]]
-      end
-      threads.each do |t|
-        if t[0] == nil
-          result.concat(thread(threads, 0, t[1], [], [], msg_id_hash))
-        end
+      forest[nil].each do |root|
+        result.concat(thread(forest, 0, root, [], [], msg_id_hash))
       end
       result
     end
 
     def one_threads(folder, folder_number)
       number = from_folder_number(folder, folder_number)
-      row = @adapter.select_one(<<'End', number)
-SELECT parent FROM disjoint WHERE number = ?
-End
-      rows = @adapter.select_all(<<'End', row[0].to_i)
-SELECT number FROM disjoint WHERE parent = ?
-End
+      sets = disjoint_sets([number])
+      forest = thread_forest(sets)
       result = Array.new
       msg_id_hash = Hash.new
-      threads = rows.collect do |row|
-        row = @adapter.select_one(<<'End', row[0].to_i)
-SELECT msg_id FROM message WHERE number = ?
-End
-        [parent_msg_id(row[0]), row[0]]
-      end
-      threads.each do |t|
-        if t[0] == nil
-          result.concat(thread(threads, 0, t[1], [], [], msg_id_hash))
-        end
+      forest[nil].each do |root|
+        result.concat(thread(forest, 0, root, [], [], msg_id_hash))
       end
       result
     end
