@@ -58,7 +58,8 @@ module MsgCab
     end
 
     def [](key)
-      Entity.new(@message.part(key), @cid + [key])
+      message = @message.part(key)
+      Entity.new(message, @cid + [key]) if message
     end
 
     def each
@@ -142,9 +143,7 @@ module MsgCab
           log(3, 'Failed to decode')
         end
       else
-        entity = self
-        entity = entity[0] while entity.multipart?
-        result = entity.decode if entity
+        result = decode
       end
       result
     end
